#ifndef LIBUVC_HX_H
#define LIBUVC_HX_H

#ifdef __cplusplus
extern "C" {
#endif

#include "libuvc/libuvc.h"

/** Himax defined macros */
#define HX_HM2056_CAM_I2C_SLAVE_ADDR_7BIT          0x24
#define HX_HM2056_CAM_I2C_WRITE_SLAVE_ADDR_8BIT    0x48
#define HX_HM2056_CAM_I2C_READ_SLAVE_ADDR_8BIT     0x49
#define HX_HV2_CAM_I2C_SLAVE_ADDR_7BIT             0x37
#define HX_HV2_CAM_I2C_WRITE_SLAVE_ADDR_8BIT       0x6E
#define HX_HV2_CAM_I2C_READ_SLAVE_ADDR_8BIT        0x6F
#define HX_HV2_FW_VERSION_STR_OFFSET               4
#define HX_HV2_FW_VERSION_STR_LEN                  8
#define HX_HV2_STREAM_MODE_GET_OFFSET              4
#define HX_HV2_ROMCODE_VERSION_STR_OFFSET          4

/** Himax frame type */
typedef enum hx_frame_type {
  HX_FRAME_TYPE_NIR = 0,
  HX_FRAME_TYPE_DEPTH,
  HX_FRAME_TYPE_DOT,
  HX_FRAME_TYPE_UNKNOWN = 0xFF,
} hx_frame_type_t;

/** Himax frame sequence mode selection ID */
typedef enum hx_frame_mode_select {
  HX_FRAME_MODE_SELECT_STREAM_MODE_0,
  HX_FRAME_MODE_SELECT_ONLY_NIR,
  HX_FRAME_MODE_SELECT_ONLY_DOT,
  HX_FRAME_MODE_SELECT_ONLY_DEPTH,
  HX_FRAME_MODE_SELECT_ALT_NIR_DEPTH,
  HX_FRAME_MODE_SELECT_ALT_NIR_DOT,
  HX_FRAME_MODE_SELECT_ALT_NIR_DEPTH_DOT,
  HX_FRAME_MODE_SELECT_AE_OFF,
  HX_FRAME_MODE_SELECT_AE_ON,
  HX_FRAME_MODE_SELECT_UNKNOWN = 0xFF,
} hx_frame_mode_select_t;

/** Himax effective bit alignment */
typedef enum hx_effective_bit_alignment {
  /** zero-padding on least-significant bits */
  HX_EFFECTIVE_BIT_ALIGNMENT_LEFT = 0,
  /** zero-padding on most-significant bits */
  HX_EFFECTIVE_BIT_ALIGNMENT_RIGHT = 1,
} hx_effective_bit_alignment_t;

/** Himax encoded RAW10 pixel data structure */
typedef struct hx_raw10_pixs {
  uint8_t B0;
  uint8_t B1;
  uint8_t B2;
  uint8_t B3;
  uint8_t B4;
} hx_raw10_pixs_t;

/** Himax ROI feature settings structure */
typedef struct hx_roi_table_feature {
  unsigned mode:2;
  unsigned type:2;
  unsigned reset:1;
  unsigned clear:1;
  unsigned update:1;
  unsigned sync:1;
} hx_roi_table_feature_t;

/** Himax NIR and RGB camera sensors' intrinsic/extrinsic parameter matrices' structure */
typedef struct hx_cam_intrinsic_extrinsic_matrices {
  /** Corresponding width/height of NIR frame size  */
  uint32_t w_d, h_d;
  /** Corresponding width/height of RGB frame size  */
  uint32_t w_r, h_r;
  /** Pre-calculated inverse matrix of
   *  NIR camera sensor intrinsic parameter matrix */
  double Kd_[3][3];
  /** NIR camera sensor intrinsic parameter matrix */
  double Kd[3][3];
  /** RGB camera sensor intrinsic parameter matrix */
  double Kr[3][3];
  /** Rotation component of extrinsic parameter matrix,
   * between NIR and RGB camera sensors */
  double R[3][3];
  /** Translation component of extrinsic parameter matrix,
   * between NIR and RGB camera sensors */
  double T[3];
} hx_cam_intrinsic_extrinsic_matrices_t;

/** Himax HV2 FW version structure */
typedef struct hx_hv2_fw_version {
  uint16_t chip_id;
  char version_major;
  uint8_t version_date_year;
  uint8_t version_date_month;
  uint8_t version_date_day;
  uint8_t version_minor;
  uint8_t customer_id;
} hx_hv2_fw_version_t;

/** Himax UVC frame rotation mode */
typedef enum uvc_hx_rotation_mode {
  /** No rotation */
  UVC_HX_ROTATION_0 = 0,
  /** Rotate 90 degrees clockwise */
  UVC_HX_ROTATION_90,
  /** Rotate 180 degrees */
  UVC_HX_ROTATION_180,
  /** Rotate 270 degrees clockwise */
  UVC_HX_ROTATION_270,
} uvc_hx_rotation_mode_t;

/** Himax UVC pixel bytes */
typedef enum uvc_hx_bpp {
  /** Grey8 Bytes per pixel */
  UVC_HX_Y8_BPP = 1,
  /** Grey16 Bytes per pixel */
  UVC_HX_Y16_BPP = 2,
  /** BGR Bytes per pixel */
  UVC_HX_BGR_BPP = 3,
  /** RGB Bytes per pixel */
  UVC_HX_RGB_BPP = 3,
} uvc_hx_bpp_t;

/** Frame */
uvc_error_t uvc_hx_y16_rotate(uvc_frame_t *in, uvc_frame_t *out, int mode);
uvc_error_t uvc_hx_bgr_rotate(uvc_frame_t *in, uvc_frame_t *out, int mode);
uvc_error_t uvc_hx_rgb_rotate(uvc_frame_t *in, uvc_frame_t *out, int mode);

uvc_error_t uvc_decode_hx_raw10_to_y16(uvc_frame_t *in, uvc_frame_t *out);
uvc_error_t uvc_y16_to_y8(uvc_frame_t *in, uvc_frame_t *out, int effective_bit_alignment);
uvc_error_t uvc_y16_shr(uvc_frame_t *in, uvc_frame_t *out, int shr);
uvc_error_t uvc_y16_shl(uvc_frame_t *in, uvc_frame_t *out, int shl);
uvc_error_t uvc_y8_to_bgr(uvc_frame_t *in, uvc_frame_t *out);
uvc_error_t uvc_y8_to_rgb(uvc_frame_t *in, uvc_frame_t *out);
uvc_error_t uvc_y16_to_depth_colormap_bgr(uvc_frame_t *in, uvc_frame_t *out, int shr);
uvc_error_t uvc_y16_to_depth_colormap_rgb(uvc_frame_t *in, uvc_frame_t *out, int shr);

/** Ctrl */
uvc_error_t uvc_hx_i2c_read_1b(uvc_device_handle_t *devh, uint8_t unit_id, uint8_t slave_addr_8bit, uint16_t reg_addr, uint8_t reg_addr_len, uint8_t *buf);
uvc_error_t uvc_hx_i2c_write_1b(uvc_device_handle_t *devh, uint8_t unit_id, uint8_t slave_addr_8bit, uint16_t reg_addr, uint8_t reg_addr_len, uint8_t buf);
uvc_error_t uvc_hx_i2c_burst_write(uvc_device_handle_t *devh, uint8_t unit_id, uint8_t slave_addr_8bit, uint16_t reg_addr, uint8_t reg_addr_len, uint8_t *buf, uint16_t buf_len);
uvc_error_t uvc_hx_i2c_burst_read(uvc_device_handle_t *devh, uint8_t unit_id, uint8_t slave_addr_8bit, uint16_t reg_addr, uint8_t reg_addr_len, uint8_t *buf, uint16_t buf_len);
uvc_error_t uvc_hx_i2c_thru_mode_read(uvc_device_handle_t *devh, uint8_t slave_addr_7bit, uint16_t reg_addr, uint8_t reg_addr_len, uint8_t *data, uint8_t data_len);
uvc_error_t uvc_hx_get_cx3_fw_version(uvc_device_handle_t *devh, char *cx3_fw_version, uint16_t cx3_fw_version_len);
uvc_error_t uvc_hx_get_ci1_fw_version(uvc_device_handle_t *devh, char *ci1_fw_version, uint16_t ci1_fw_version_len);
uvc_error_t uvc_hx_get_hv2_fw_version(uvc_device_handle_t *devh, hx_hv2_fw_version_t *hv2_fw_version);
uvc_error_t uvc_hx_get_hv2_romcode_version(uvc_device_handle_t *devh, uint8_t *hv2_romcode_version, uint8_t hv2_romcode_version_len);
uvc_error_t uvc_hx_get_frame_mode(uvc_device_handle_t *devh, uint8_t *mode_id);
uvc_error_t uvc_hx_set_frame_mode(uvc_device_handle_t *devh, enum hx_frame_mode_select mode_id);
uvc_error_t uvc_hx_roi_set_feature(uvc_device_handle_t *devh, struct hx_roi_table_feature *feature);
uvc_error_t uvc_hx_roi_set_coordinate(uvc_device_handle_t *devh, uint8_t weighting, int *box, int size);
uvc_error_t uvc_hx_set_2d_target_fmean(uvc_device_handle_t *devh, uint16_t fmean, uint16_t bound);
uvc_error_t uvc_hx_get_2d_target_fmean(uvc_device_handle_t *devh, uint16_t *fmean, uint16_t *bound);
uvc_error_t uvc_hx_set_again(uvc_device_handle_t *devh, uint8_t again);
uvc_error_t uvc_hx_get_build_type(uvc_device_handle_t *devh, uint16_t *build_type);
uvc_error_t uvc_hx_get_serial_number(uvc_device_handle_t *devh, uint32_t *serial_number);
uvc_error_t uvc_hx_set_face_recognize(uvc_device_handle_t *devh, uint8_t mode, uint16_t face_id, uint16_t *result);
uvc_error_t uvc_hx_set_liveness_check(uvc_device_handle_t *devh, uint8_t mode, uint16_t *result);
uvc_error_t uvc_hx_get_intrinsic_extrinsic_matrices(uvc_device_handle_t *devh, hx_cam_intrinsic_extrinsic_matrices_t *cam_matrices, int require_scaling);
uvc_error_t uvc_hx_flash_cmd(uvc_device_handle_t *devh, const char *filename, uint8_t type);

/** Stream */
uvc_error_t uvc_get_stream_ctrl_format_size_ifno(
    uvc_device_handle_t *devh,
    uvc_stream_ctrl_t *ctrl,
    enum uvc_frame_format format,
    int width, int height,
    int fps,
    uint8_t ifno
    );

uvc_error_t uvc_start_streaming_multi_cb(
    uvc_device_handle_t *devh,
    uvc_stream_ctrl_t *ctrl,
    uvc_frame_callback_t *nir_cb,
    void *nir_user_ptr,
    uvc_frame_callback_t *depth_cb,
    void *depth_user_ptr,
    uvc_frame_callback_t *dot_cb,
    void *dot_user_ptr,
    uint8_t flags);

uvc_error_t uvc_stream_start_multi_cb(uvc_stream_handle_t *strmh,
    uvc_frame_callback_t *nir_cb,
    void *nir_user_ptr,
    uvc_frame_callback_t *depth_cb,
    void *depth_user_ptr,
    uvc_frame_callback_t *dot_cb,
    void *dot_user_ptr,
    uint8_t flags);

/** Misc */
/**
 * @brief Himax NIR to RGB camera sensor coordinate converter of portrait image.
 * @ingroup misc
 *
 * @param[in] cam_matrices Pointer to Himax NIR and RGB camera sensors'
 *                         intrinsic/extrinsic parameter matrices' structure.
 * @param[in] depth Depth value at a NIR point (x_nir, y_nir), unit: 1/16 mm (Q12.4).
 * @param[in] x_nir X of a NIR point.
 * @param[in] y_nir Y of a NIR point.
 * @param[out] x_rgb X of a theoretical RGB point.
 * @param[out] y_rgb Y of a theoretical RGB point.
 * @return On success, returned (0).  Otherwise,
 *   non-zero value describing the error encountered.
 */
int uvc_hx_rectified_coordinate(
    const hx_cam_intrinsic_extrinsic_matrices_t *cam_matrices,
    uint16_t depth,
    uint32_t x_nir,
    uint32_t y_nir,
    uint32_t *x_rgb,
    uint32_t *y_rgb);

/** Diag */
uvc_error_t uvc_hx_fprintf_intrinsic_extrinsic_matrices(FILE *stream, hx_cam_intrinsic_extrinsic_matrices_t *cam_matrices);

#ifdef __cplusplus
}
#endif

#endif // !def(LIBUVC_HX_H)
