/*********************************************************************
* Software License Agreement (BSD License)
*
*  Copyright (C) 2010-2012 Ken Tossell
*  All rights reserved.
*
*  Redistribution and use in source and binary forms, with or without
*  modification, are permitted provided that the following conditions
*  are met:
*
*   * Redistributions of source code must retain the above copyright
*     notice, this list of conditions and the following disclaimer.
*   * Redistributions in binary form must reproduce the above
*     copyright notice, this list of conditions and the following
*     disclaimer in the documentation and/or other materials provided
*     with the distribution.
*   * Neither the name of the author nor other contributors may be
*     used to endorse or promote products derived from this software
*     without specific prior written permission.
*
*  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
*  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
*  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
*  FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
*  COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
*  INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
*  BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
*  LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
*  CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
*  LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
*  ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
*  POSSIBILITY OF SUCH DAMAGE.
*********************************************************************/

/**
 * @defgroup test Example code of test application
 * @brief Capture video frame of UVC device, and render in OpenCV window
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <unistd.h>
#include <X11/Xlib.h>
#include <opencv2/highgui/highgui_c.h>

#include "libuvc/libuvc_hx.h"

#define USB_VID         0x2AAD
#define USB_PID         0x6373

#define ENABLE_NIR_SENSOR 1
#if ENABLE_NIR_SENSOR
#define CAPTURE_WIDTH   640
#define CAPTURE_HEIGHT  401
#define CAPTURE_FPS     30
#define CAPTURE_WIDTH_2 1280
#define CAPTURE_HEIGHT_2 801
#define CAPTURE_FPS_2   30
#define CAPTURE_FORMAT  UVC_FRAME_FORMAT_YUYV
#endif  // ENABLE_NIR_SENSOR

#define ENABLE_RGB_SENSOR   1
#if ENABLE_RGB_SENSOR
#define RGB_CAPTURE_WIDTH   640
#define RGB_CAPTURE_HEIGHT  481
#define RGB_CAPTURE_FPS     30
#define RGB_CAPTURE_WIDTH_2 1280
#define RGB_CAPTURE_HEIGHT_2 801
#define RGB_CAPTURE_FPS_2   15
#define RGB_CAPTURE_FORMAT  UVC_FRAME_FORMAT_YUYV
#endif  // ENABLE_RGB_SENSOR

#define ENABLE_DUMP_IMAGE      0
#define DUMP_IMAGE_MAX_COUNT   10

#define CI1_FW_VERSION_LEN     32

/*  Macros to cut down on compiler warnings. */
#define UNUSED(expr) do { (void)(expr); } while (0)

static int non_stop = 1;
static int resolution_selection = 1;

/**
 * @brief Dump frame buffer to file.
 * @ingroup debug
 */
#if ENABLE_DUMP_IMAGE
void save_img(const void *buf, size_t len, const char *fname_prefix, int fname_idx)
{
  char fname[100] = {0};
  FILE *fp = NULL;

  snprintf(fname, 100, "%s_%02d.raw", fname_prefix, fname_idx);
  if ((fp = fopen(fname, "w+")) != NULL) {
    fwrite(buf, len, 1, fp);
    fclose(fp);
    printf("Saved image: %s\n", fname);
  }
}
#else  // !ENABLE_DUMP_IMAGE
void save_img(const void *buf, size_t len, const char *fname_prefix, int fname_idx)
{
  UNUSED(buf); UNUSED(len); UNUSED(fname_prefix); UNUSED(fname_idx);
}
#endif  // ENABLE_DUMP_IMAGE

/**
 * @brief SIGINT signal handler.
 * @ingroup test
 */
void sigint(int a) {
  UNUSED(a);
  printf("\naborting...\n");
  non_stop = 0;
}

#if ENABLE_NIR_SENSOR
/**
 * @brief NIR frame callback function.
 * @ingroup test
 */
void nir_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *y16_r270;
  uvc_frame_t *y10_r270;
  uvc_frame_t *y8_r270;
  uvc_frame_t *bgr_r270;
  uvc_error_t  ret;
  IplImage *cvImg;
  static int cnt_nir_raw = 1;
  static int cnt_nir_y16_r270 = 1;
  static int cnt_nir_y10_r270 = 1;
  static int cnt_nir_y8_r270 = 1;
  static int cnt_nir_bgr888_r270 = 1;

  UNUSED(ptr);

  if (!frame) {
    printf("NIR frame nullptr!\n");
    return;
  }

  if (cnt_nir_raw <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_nir_raw", cnt_nir_raw++);

  y16_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y16_r270) {
    printf("NIR: unable to allocate y16_r270 frame!\n");
    return;
  }

  y10_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y10_r270) {
    printf("NIR: unable to allocate y10_r270 frame!\n");
    uvc_free_frame(y16_r270);
    return;
  }

  y8_r270 = uvc_allocate_frame(frame->width * frame->height * 1);
  if (!y8_r270) {
    printf("NIR: unable to allocate y8_r270 frame!\n");
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270) {
    printf("NIR: unable to allocate bgr_r270 frame!\n");
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  ret = uvc_hx_y16_rotate(frame, y16_r270, UVC_HX_ROTATION_270);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "NIR: uvc_hx_y16_rotate(270)");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_nir_y16_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y16_r270->data, y16_r270->data_bytes, "img_nir_y16_r270", cnt_nir_y16_r270++);

  // Left-aligned => right-aligned 10-bit
  ret = uvc_y16_shr(y16_r270, y10_r270, 6);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "NIR: uvc_y16_shr(6)");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_nir_y10_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y10_r270->data, y10_r270->data_bytes, "img_nir_y10_r270", cnt_nir_y10_r270++);

  ret = uvc_y16_to_y8(y16_r270, y8_r270, HX_EFFECTIVE_BIT_ALIGNMENT_LEFT);
  if (ret) {
    uvc_perror(ret, "NIR: uvc_y16_to_y8");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_nir_y8_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y8_r270->data, y8_r270->data_bytes, "img_nir_y8_r270", cnt_nir_y8_r270++);

  ret = uvc_any2bgr(y8_r270, bgr_r270);
  if (ret) {
    uvc_perror(ret, "NIR: uvc_any2bgr");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_nir_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_nir_bgr888_r270", cnt_nir_bgr888_r270++);

  cvImg = cvCreateImageHeader(cvSize(bgr_r270->width, bgr_r270->height), IPL_DEPTH_8U, 3);
  cvSetData(cvImg, bgr_r270->data, bgr_r270->width * 3);
  cvNamedWindow("NIR", CV_WINDOW_AUTOSIZE);
  cvShowImage("NIR", cvImg);
  cvWaitKey(10);
  cvReleaseImageHeader(&cvImg);
  uvc_free_frame(bgr_r270);
  uvc_free_frame(y8_r270);
  uvc_free_frame(y10_r270);
  uvc_free_frame(y16_r270);
}
#endif  // ENABLE_NIR_SENSOR

#if ENABLE_NIR_SENSOR
/**
 * @brief Depth frame callback function.
 * @ingroup test
 */
void depth_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *y16_r270;
  uvc_frame_t *q12_r270;
  uvc_frame_t *bgr_r270;
  uvc_error_t  ret;
  IplImage *cvImg;
  static int cnt_dep_raw = 1;
  static int cnt_dep_y16_r270 = 1;
  static int cnt_dep_q12_r270 = 1;
  static int cnt_dep_bgr888_r270 = 1;

  UNUSED(ptr);

  if (!frame) {
    printf("Depth frame nullptr!\n");
    return;
  }

  if (cnt_dep_raw <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_dep_raw", cnt_dep_raw++);

  y16_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y16_r270) {
    printf("Depth: unable to allocate y16_r270 frame!\n");
    return;
  }

  q12_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!q12_r270) {
    printf("Depth: unable to allocate q12_r270 frame!\n");
    uvc_free_frame(y16_r270);
    return;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270) {
    printf("Depth: unable to allocate bgr_r270 frame!\n");
    uvc_free_frame(q12_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  ret = uvc_hx_y16_rotate(frame, y16_r270, UVC_HX_ROTATION_270);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "Depth: uvc_hx_y16_rotate(270)\n");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(q12_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dep_y16_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y16_r270->data, y16_r270->data_bytes, "img_dep_y16_r270", cnt_dep_y16_r270++);

  // Q12.4 => Q12
  ret = uvc_y16_shr(y16_r270, q12_r270, 4);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "Depth: uvc_y16_shr(4)");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(q12_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dep_q12_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(q12_r270->data, q12_r270->data_bytes, "img_dep_q12_r270", cnt_dep_q12_r270++);

  ret = uvc_y16_to_depth_colormap_bgr(q12_r270, bgr_r270, 1);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "Depth: uvc_y16_to_depth_colormap_bgr");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(q12_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dep_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_dep_bgr888_r270", cnt_dep_bgr888_r270++);

  cvImg = cvCreateImageHeader(cvSize(bgr_r270->width, bgr_r270->height), IPL_DEPTH_8U, 3);
  cvSetData(cvImg, bgr_r270->data, bgr_r270->width * 3);
  cvNamedWindow("Depth", CV_WINDOW_AUTOSIZE);
  cvShowImage("Depth", cvImg);
  cvWaitKey(10);
  cvReleaseImageHeader(&cvImg);
  uvc_free_frame(bgr_r270);
  uvc_free_frame(q12_r270);
  uvc_free_frame(y16_r270);
}
#endif  // ENABLE_NIR_SENSOR

#if ENABLE_NIR_SENSOR
/**
 * @brief Dot frame callback function.
 * @ingroup test
 */
void dot_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *y16_r270;
  uvc_frame_t *y10_r270;
  uvc_frame_t *y8_r270;
  uvc_frame_t *bgr_r270;
  uvc_error_t  ret;
  IplImage *cvImg;
  static int cnt_dot_raw = 1;
  static int cnt_dot_y16_r270 = 1;
  static int cnt_dot_y10_r270 = 1;
  static int cnt_dot_y8_r270 = 1;
  static int cnt_dot_bgr888_r270 = 1;

  UNUSED(ptr);

  if (!frame) {
    printf("Dot frame nullptr!\n");
    return;
  }

  if (cnt_dot_raw <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_dot_raw", cnt_dot_raw++);

  y16_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y16_r270) {
    printf("Dot: unable to allocate y16_r270 frame!\n");
    return;
  }

  y10_r270 = uvc_allocate_frame(frame->width * frame->height * 2);
  if (!y10_r270) {
    printf("Dot: unable to allocate y10_r270 frame!\n");
    uvc_free_frame(y16_r270);
    return;
  }

  y8_r270 = uvc_allocate_frame(frame->width * frame->height * 1);
  if (!y8_r270) {
    printf("Dot: unable to allocate y8_r270 frame!\n");
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270) {
    printf("Dot: unable to allocate bgr_r270 frame!\n");
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  ret = uvc_hx_y16_rotate(frame, y16_r270, UVC_HX_ROTATION_270);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "Dot: uvc_hx_y16_rotate(270)\n");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dot_y16_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y16_r270->data, y16_r270->data_bytes, "img_dot_y16_r270", cnt_dot_y16_r270++);

  // Left-aligned => right-aligned 10-bit
  ret = uvc_y16_shr(y16_r270, y10_r270, 6);
  if (ret != UVC_SUCCESS) {
    uvc_perror(ret, "Dot: uvc_y16_shr(6)");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dot_y10_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y10_r270->data, y10_r270->data_bytes, "img_dot_y10_r270", cnt_dot_y10_r270++);

  ret = uvc_y16_to_y8(y16_r270, y8_r270, HX_EFFECTIVE_BIT_ALIGNMENT_LEFT);
  if (ret) {
    uvc_perror(ret, "Dot: uvc_y16_to_y8");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dot_y8_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(y8_r270->data, y8_r270->data_bytes, "img_dot_y8_r270", cnt_dot_y8_r270++);

  ret = uvc_any2bgr(y8_r270, bgr_r270);
  if (ret) {
    uvc_perror(ret, "Dot: uvc_any2bgr");
    uvc_free_frame(bgr_r270);
    uvc_free_frame(y8_r270);
    uvc_free_frame(y10_r270);
    uvc_free_frame(y16_r270);
    return;
  }

  if (cnt_dot_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_dot_bgr888_r270", cnt_dot_bgr888_r270++);

  cvImg = cvCreateImageHeader(cvSize(bgr_r270->width, bgr_r270->height), IPL_DEPTH_8U, 3);
  cvSetData(cvImg, bgr_r270->data, bgr_r270->width * 3);
  cvNamedWindow("Dot", CV_WINDOW_AUTOSIZE);
  cvShowImage("Dot", cvImg);
  cvWaitKey(10);
  cvReleaseImageHeader(&cvImg);
  uvc_free_frame(bgr_r270);
  uvc_free_frame(y8_r270);
  uvc_free_frame(y10_r270);
  uvc_free_frame(y16_r270);
}
#endif  // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
/**
 * @brief RGB Frame callback function.
 * @ingroup test
 */
void rgb_cb(uvc_frame_t *frame, void *ptr)
{
  uvc_frame_t *bgr;
  uvc_frame_t *bgr_r270;
  uvc_error_t  ret;
  IplImage *cvImg;
  static int cnt_rgb_yuv422 = 1;
  static int cnt_rgb_bgr888 = 1;
  static int cnt_rgb_bgr888_r270 = 1;

  UNUSED(ptr);

  if (!frame) {
    printf("RGB frame nullptr!\n");
    return;
  }

  if (cnt_rgb_yuv422 <= DUMP_IMAGE_MAX_COUNT)
    save_img(frame->data, frame->data_bytes, "img_rgb_yuv422", cnt_rgb_yuv422++);

  bgr = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr) {
    printf("RGB: unable to allocate bgr frame!\n");
    return;
  }

  bgr_r270 = uvc_allocate_frame(frame->width * frame->height * 3);
  if (!bgr_r270) {
    printf("Dot: unable to allocate bgr_r270 frame!\n");
    uvc_free_frame(bgr);
    return;
  }

  ret = uvc_any2bgr(frame, bgr);
  if (ret) {
    uvc_perror(ret, "RGB: uvc_any2bgr");
    uvc_free_frame(bgr);
    uvc_free_frame(bgr_r270);
    return;
  }

  if (cnt_rgb_bgr888 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr->data, bgr->data_bytes, "img_rgb_bgr888", cnt_rgb_bgr888++);

  ret = uvc_hx_bgr_rotate(bgr, bgr_r270, UVC_HX_ROTATION_270);
  if (ret) {
    uvc_perror(ret, "RGB: uvc_hx_rgb_rotate(270)\n");
    uvc_free_frame(bgr);
    uvc_free_frame(bgr_r270);
    return;
  }

  if (cnt_rgb_bgr888_r270 <= DUMP_IMAGE_MAX_COUNT)
    save_img(bgr_r270->data, bgr_r270->data_bytes, "img_rgb_bgr888_r270", cnt_rgb_bgr888_r270++);

  cvImg = cvCreateImageHeader(cvSize(bgr_r270->width, bgr_r270->height), IPL_DEPTH_8U, 3);
  cvSetData(cvImg, bgr_r270->data, bgr_r270->width * 3);
  cvNamedWindow("RGB", CV_WINDOW_AUTOSIZE);
  cvShowImage("RGB", cvImg);
  cvWaitKey(10);
  cvReleaseImageHeader(&cvImg);
  uvc_free_frame(bgr);
  uvc_free_frame(bgr_r270);
}
#endif  // ENABLE_RGB_SENSOR

/**
 * @brief main function.
 * @ingroup test
 */
int main(int argc, char **argv) {
  uvc_context_t *ctx;
  uvc_error_t res;
  uvc_device_t *dev;
  uvc_device_handle_t *devh;
#if ENABLE_NIR_SENSOR
  uvc_stream_ctrl_t ctrl;
#endif  // ENABLE_NIR_SENSOR
  char ci1_fw_version[CI1_FW_VERSION_LEN] = { 0 };
  hx_hv2_fw_version_t hv2_fw_version;
  uint8_t hv2_romcode_version[4];
  uint8_t current_stream_mode;
#if ENABLE_RGB_SENSOR
  uvc_error_t rgb_res;
  uvc_stream_ctrl_t rgb_ctrl;
#endif  // ENABLE_RGB_SENSOR
  hx_cam_intrinsic_extrinsic_matrices_t cam_matrices;
  uint16_t build_type;
  uint32_t serial_number;

  UNUSED(argc);
  UNUSED(argv);

  XInitThreads();

  res = uvc_init(&ctx, NULL);

  if (res < 0) {
    uvc_perror(res, "uvc_init");
    return res;
  }

  puts("UVC initialized");

  res = uvc_find_device(ctx, &dev, USB_VID, USB_PID, NULL);

  if (res < 0) {
    uvc_perror(res, "uvc_find_device");
  } else {
    puts("Device found");

    res = uvc_open(dev, &devh);

    if (res < 0) {
      uvc_perror(res, "uvc_open");
    } else {
      puts("Device opened");

      //uvc_print_diag(devh, stderr);

      res = uvc_hx_get_ci1_fw_version(devh, ci1_fw_version, sizeof(ci1_fw_version));
      if (res != UVC_SUCCESS) {
        uvc_perror(res, "uvc_hx_get_ci1_fw_version failed");
      }
      printf("ci1_fw_version = %s\n", ci1_fw_version);

      res = uvc_hx_get_hv2_fw_version(devh, &hv2_fw_version);
      if (res != UVC_SUCCESS) {
        uvc_perror(res, "uvc_hx_get_hv2_fw_version failed");
      }
      printf("hv2_fw_version: chip_id=0x%04X, version_major='%c', "
             "version_date=%02u.%02u.%02u, version_minor=0x%02X, customer_id=0x%02X\n",
          hv2_fw_version.chip_id,
          hv2_fw_version.version_major,
          hv2_fw_version.version_date_year,
          hv2_fw_version.version_date_month,
          hv2_fw_version.version_date_day,
          hv2_fw_version.version_minor,
          hv2_fw_version.customer_id);

      res = uvc_hx_get_hv2_romcode_version(devh, hv2_romcode_version, sizeof(hv2_romcode_version));
      if (res != UVC_SUCCESS) {
        uvc_perror(res, "uvc_hx_get_hv2_romcode_version failed");
      } else {
        printf("hv2_romcode_version: %02X %02X %02X %02X\n",
            hv2_romcode_version[0],
            hv2_romcode_version[1],
            hv2_romcode_version[2],
            hv2_romcode_version[3]);
      }

      build_type = 0;
      res = uvc_hx_get_build_type(devh, &build_type);
      if (res != UVC_SUCCESS) {
        uvc_perror(res, "Failed to get Himax camera build type");
      } else {
        printf("Himax camera build type = 0x%04X\n", build_type);
      }

      serial_number = 0;
      res = uvc_hx_get_serial_number(devh, &serial_number);
      if (res != UVC_SUCCESS) {
        uvc_perror(res, "Failed to get Himax camera serial number");
      } else {
        printf("Himax camera serial number = 0x%08X (%u)\n", serial_number, serial_number);
      }

      memset(&cam_matrices, 0, sizeof(hx_cam_intrinsic_extrinsic_matrices_t));
      res = uvc_hx_get_intrinsic_extrinsic_matrices(devh, &cam_matrices, 0);
      if (res != UVC_SUCCESS) {
        uvc_perror(res, "Failed to get Himax camera sensors' intrinsic extrinsic parameter matrices");
      } else {
        printf("Himax camera sensors' intrinsic extrinsic parameter matrices:\n");
        res = uvc_hx_fprintf_intrinsic_extrinsic_matrices(stdout, &cam_matrices);
        if (res != UVC_SUCCESS) {
          printf("Failed to print Himax camera sensors' intrinsic extrinsic parameter matrices\n");
        }
      }
#if ENABLE_NIR_SENSOR || ENABLE_RGB_SENSOR

#if ENABLE_NIR_SENSOR
      res = uvc_get_stream_ctrl_format_size_ifno(
        devh, &ctrl, CAPTURE_FORMAT, CAPTURE_WIDTH_2, CAPTURE_HEIGHT_2, CAPTURE_FPS_2, 1);

      uvc_print_stream_ctrl(&ctrl, stderr);
#endif  // ENABLE_NIR_SENSOR

#if ENABLE_RGB_SENSOR
      rgb_res = uvc_get_stream_ctrl_format_size_ifno(
        devh, &rgb_ctrl, RGB_CAPTURE_FORMAT, RGB_CAPTURE_WIDTH_2, RGB_CAPTURE_HEIGHT_2, RGB_CAPTURE_FPS_2, 2);

      uvc_print_stream_ctrl(&rgb_ctrl, stderr);
#endif  // ENABLE_RGB_SENSOR

#if ENABLE_NIR_SENSOR
      if (res < 0) {
        uvc_perror(res, "get_mode");
#else  // !ENABLE_NIR_SENSOR
      if (rgb_res < 0) {
#endif  // ENABLE_NIR_SENSOR
#if ENABLE_RGB_SENSOR
      } else if (rgb_res < 0) {
        uvc_perror(res, "RGB get_mode");
#endif  // ENABLE_RGB_SENSOR
      } else {
#if ENABLE_NIR_SENSOR
        res = uvc_start_streaming_multi_cb(devh, &ctrl,
            nir_cb, (void *) 1001,
            depth_cb, (void *) 1002,
            dot_cb, (void *) 1003,
            0);
#endif  // ENABLE_NIR_SENSOR
#if ENABLE_RGB_SENSOR
        rgb_res = uvc_start_streaming(devh, &rgb_ctrl, rgb_cb, (void *) 67890, 0);
#endif  // ENABLE_RGB_SENSOR

#if ENABLE_NIR_SENSOR
        if (res < 0) {
          uvc_perror(res, "start_streaming");
#else  // !ENABLE_NIR_SENSOR
        if (rgb_res < 0) {
#endif  // ENABLE_NIR_SENSOR
#if ENABLE_RGB_SENSOR
        } else if (rgb_res < 0) {
          uvc_perror(res, "RGB start_streaming");
#endif  // ENABLE_RGB_SENSOR
        } else {
          //res = uvc_hx_set_frame_mode(devh, HX_FRAME_MODE_SELECT_ONLY_NIR);
          //res = uvc_hx_set_frame_mode(devh, HX_FRAME_MODE_SELECT_ONLY_DOT);
          //res = uvc_hx_set_frame_mode(devh, HX_FRAME_MODE_SELECT_ONLY_DEPTH);
          res = uvc_hx_set_frame_mode(devh, HX_FRAME_MODE_SELECT_ALT_NIR_DEPTH);
          //res = uvc_hx_set_frame_mode(devh, HX_FRAME_MODE_SELECT_ALT_NIR_DOT);
          //res = uvc_hx_set_frame_mode(devh, HX_FRAME_MODE_SELECT_ALT_NIR_DEPTH_DOT);
          if (res != UVC_SUCCESS) {
            uvc_perror(res, "uvc_hx_set_frame_mode failed");
          }

          /** Send a Face recognize command.
           * uvc_error_t frc_ret = UVC_SUCCESS;
           * uint16_t frc_result = 0;
           * frc_ret = uvc_hx_set_face_recognize(devh, 0, 0x0001, &frc_result);
           * if (frc_ret != UVC_SUCCESS) {
           *   uvc_perror(res, "uvc_hx_set_face_recognize(0) failed");
           * } else {
           *   printf("Returned face recognize result = 0x%04X\n", frc_result);
           * }
           */

          /** Send a Face register command with target Face ID = 1.
           * uvc_error_t frg_ret = UVC_SUCCESS;
           * uint16_t frg_result = 0;
           * frg_ret = uvc_hx_set_face_recognize(devh, 1, 0x0001, &frg_result);
           * if (frg_ret != UVC_SUCCESS) {
           *   uvc_perror(res, "uvc_hx_set_face_recognize(1) failed");
           * } else {
           *   printf("Returned face register result = 0x%04X\n", frg_result);
           * }
           */

          /** Send a Liveness check command.
           * uvc_error_t flc_ret = UVC_SUCCESS;
           * uint16_t flc_result = 0;
           * flc_ret = uvc_hx_set_liveness_check(devh, 0, &flc_result);
           * if (flc_ret != UVC_SUCCESS) {
           *   uvc_perror(res, "uvc_hx_set_liveness_check failed");
           * } else {
           *   printf("Returned liveness check result = 0x%04X\n", flc_result);
           * }
           */

          res = uvc_hx_get_frame_mode(devh, &current_stream_mode);
          if (res != UVC_SUCCESS) {
            uvc_perror(res, "uvc_hx_get_frame_mode failed");
          } else {
            printf("current stream mode = %u\n", current_stream_mode);
          }

          puts("Streaming... wait for pressing [Ctrl-C] to terminate");
          signal(SIGINT, sigint);
          while (non_stop) {
            sleep(5);
            uvc_stop_streaming(devh);
            if (resolution_selection == 0) {
#if ENABLE_NIR_SENSOR
              uvc_get_stream_ctrl_format_size_ifno(
                devh, &ctrl, CAPTURE_FORMAT, CAPTURE_WIDTH, CAPTURE_HEIGHT, CAPTURE_FPS, 1);
#endif  // ENABLE_NIR_SENSOR
#if ENABLE_RGB_SENSOR
              uvc_get_stream_ctrl_format_size_ifno(
                devh, &rgb_ctrl, RGB_CAPTURE_FORMAT, RGB_CAPTURE_WIDTH, RGB_CAPTURE_HEIGHT, RGB_CAPTURE_FPS, 2);
#endif  // ENABLE_RGB_SENSOR
            } else {
#if ENABLE_NIR_SENSOR
              uvc_get_stream_ctrl_format_size_ifno(
                devh, &ctrl, CAPTURE_FORMAT, CAPTURE_WIDTH_2, CAPTURE_HEIGHT_2, CAPTURE_FPS_2, 1);
#endif  // ENABLE_NIR_SENSOR
#if ENABLE_RGB_SENSOR
              uvc_get_stream_ctrl_format_size_ifno(
                devh, &rgb_ctrl, RGB_CAPTURE_FORMAT, RGB_CAPTURE_WIDTH_2, RGB_CAPTURE_HEIGHT_2, RGB_CAPTURE_FPS_2, 2);
#endif  // ENABLE_RGB_SENSOR
            }
#if ENABLE_NIR_SENSOR
            uvc_start_streaming_multi_cb(devh, &ctrl,
                nir_cb, (void *) 1001,
                depth_cb, (void *) 1002,
                dot_cb, (void *) 1003,
                0);
#endif  // ENABLE_NIR_SENSOR
#if ENABLE_RGB_SENSOR
            uvc_start_streaming(devh, &rgb_ctrl, rgb_cb, (void *) 67890, 0);
#endif  // ENABLE_RGB_SENSOR
            resolution_selection = (resolution_selection + 1) % 2;

            res = uvc_hx_get_frame_mode(devh, &current_stream_mode);
            if (res != UVC_SUCCESS) {
              uvc_perror(res, "uvc_hx_get_frame_mode failed");
            } else {
              printf("current stream mode = %u\n", current_stream_mode);
            }
          }
          uvc_stop_streaming(devh);
          puts("Done streaming.");
        }
      }
#endif  // ENABLE_NIR_SENSOR || ENABLE_RGB_SENSOR

      uvc_close(devh);
      puts("Device closed");

    }

    uvc_unref_device(dev);
  }

  uvc_exit(ctx);
  puts("UVC exited");

  return 0;
}

